/**** << Text Formating & Printing Utilities >> ****
 *
 *  cmt.c	Ver 2.00
 *                                                  Oct 21 1982
 *                                      mod:        Mar 22 1994
 ************************************************** tonooka ***********/
/*
 *	Copyright (c) 1994 Yasuhiro Tonooka (tonooka@msi.co.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#if !defined lint
static char *copyright = "@(#)Copyright (c) 1994 Yasuhiro Tonooka";
static char *sccsid = "@(#)cmt.c 2.2 (Y.Tonooka) 3/28/94";
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "cmt.h"
#include "euc.h"

#define BOLDFACE	(1 << 0)
#define UNDERLINE	(1 << 1)

#define TABLEALLOC	64

/*
 * Regular expression routines.
 */
extern char_t *regmatch();	/* match a string to an expression */
extern union regexp *regcomp();	/* Compiles regular expression to
				 * internal form */
extern void regfree();		/* free regular expression */

/*
 * "cmtdefs" routines.
 */
extern long getentry();
extern char_t *getname();
extern char_t *getstr();
extern bool getflag();

/*
 * Std library routines from "string.h".
 */
extern char *strcat();
extern int strcmp(), strncmp(), strcasecmp();

static char *deffile;		/* name of language definition file */

/**********************************************************************
 *                                                                    *
 *  Main Routines                                                     *
 *                                                                    *
 **********************************************************************/
static char *progname;		/* program name */

static bool keyword = 1;	/* keyword switch */
static bool onecase = 0;	/* case insensitivity flag */
static bool verbose = 0;
static char_t *addkwd = NULL;	/* additional keyword buffer */

enum nametype { FNAME, LNAME };
static long deflang();
static void makelangtable();
static char_t *readkeyword();
static void process();
static union regexp *compile();

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    NAME
	main
 ---------------------------------------------------------------------*/
main(c, v)
    int c;
    register char *v[];
{
    char *langname = NULL;	/* used by -l & -L option */
    enum nametype nametype;	/* FNAME or LNAME */
    long language = -1L;	/* specified by -l or -L */
    long lang;
    char *filename;
    int status;
    int dev, ino = -1;
    struct stat sbuf;
    FILE *iop;
    FILE *defp;
    char *s;
    char_t *ss;
    extern char *ttyname();

    progname = *v;
    /*
     * Process options here.
     */
    for (v++; --c; v++) {
	/*
	 * Specify the language with -lxx.
	 */
	if (strncmp(*v, "-l", 2) == 0) {
	    if (v[0][2] == '\0')
		error("-l: bad language specification");
	    langname = *v + 2;
	    nametype = LNAME;
	    continue;
	}
	/*
	 * Specify the language with -L xx, where xx stands for
	 * filename.  This option is useful when a file is filtered
	 * prior to this command. 
	 */
	if (strcmp(*v, "-L") == 0) {
	    if (--c == 0)
		error("%s option must have an argument", *v);
	    v++;
	    langname = *v;
	    nametype = FNAME;
	    continue;
	}
	/*
	 * Specify the language definition file.
	 */
	if (strcmp(*v, "-d") == 0) {
	    if (--c == 0)
		error("%s option must have an argument", *v);
	    v++;
	    deffile = *v;
	    continue;
	}
	/*
	 * Additional keyword file with -f option.
	 */
	if (strcmp(*v, "-f") == 0) {
	    if (--c == 0)
		error("%s option must have an argument", *v);
	    v++;
	    if (addkwd)
		error("duplicate specification of -f option");
	    addkwd = readkeyword(*v);
	    continue;
	}
	/*
	 * With -, cmt takes its input from stdin.
	 */
	if ((*v)[0] != '-' || !(*v)[1])
	    break;
	for (s = *v + 1; *s; s++) {
	    switch (*s) {
	    case 'k':
		/*
		 * -k option disables making keyword boldface.
		 */
		keyword = 0;
		break;
	    case 'i':
		/*
		 * Case insensitive (one case) flag.
		 */
		onecase = 1;
		break;
	    case 'v':
		/*
		 * Tell the language name.
		 */
		verbose = 1;
		break;
	    default:
		error("-%c: bad option", (int) *s);
	    }
	}
    }
    /*
     * Prepare definition file name.
     */
    if (deffile == NULL) {
	if ((deffile = (char *) malloc(sizeof (LIBPATH)
		+ sizeof (DEFSFILE) + 2)) == NULL)
	    error("out of memory");
	sprintf(deffile, "%s/%s", LIBPATH, DEFSFILE);
    }
    /*
     * Open "cmtdefs".
     */
    if ((defp = fopen(deffile, "r")) == NULL) {
	perror(deffile);
	exit(1);
    }
    /*
     * Make language table.
     */
    makelangtable(defp);
    if (langname)
	if ((language = deflang(langname, nametype)) < 0)
	    error("language type unknown for %s", langname);
    /*
     * Get some info on output file.
     */
    if (fstat(fileno(stdout), &sbuf) == 0) {
	sbuf.st_mode &= S_IFMT;
	if (sbuf.st_mode != S_IFCHR && sbuf.st_mode != S_IFBLK) {
	    dev = sbuf.st_dev;
	    ino = sbuf.st_ino;
	}
    }
    /*
     * Main loop.
     */
    status = 0;
    do {
	if (c == 0 || strcmp(*v, "-") == 0) {
	    /*
	     * Stdin:  If tty and at EOF, clear EOF.
	     */
	    if (feof(stdin))
		if (isatty(fileno(stdin)))
		    clearerr(stdin);
		else
		    continue;
	    iop = stdin;
	    filename = NULL;
	} else {
	    /*
	     * Open a file.
	     */
	    if ((iop = fopen(*v, "r")) == NULL) {
		perror(*v);
		status |= 1;
		continue;
	    }
	    filename = *v;
	}
	if (fstat(fileno(iop), &sbuf) == 0) {
	    /*
	     * Get some info on input file, and see if it is a
	     * directory. 
	     */
	    if ((sbuf.st_mode & S_IFMT) == S_IFDIR) {
		fprintf(stderr,
			"%s: read error on %s: Is a directory\n",
			progname,
			filename ? filename : "standard input");
		fclose(iop);
		status |= 1;
		continue;
	    }
	    /*
	     * Compare the info of input with that of output, and see
	     * if they are identical. 
	     */
	    if ((sbuf.st_mode & S_IFMT) == S_IFREG &&
		    sbuf.st_dev == dev && sbuf.st_ino == ino) {
		fprintf(stderr, "%s: input %s is output\n", progname,
			filename ? filename : "-");
		fclose(iop);
		status |= 1;
		continue;
	    }
	}
	/*
	 * Get the definition for the language from the "cmtdefs" file.
	 */
	lang = language >= 0 ? language
		: deflang(filename ? filename : "-", FNAME);
	if (lang < 0)
	    error("language type unknown for %s",
		    filename ? filename : "standard input");
	if (getentry(defp, lang) < 0)
	    error("something's wrong: cannot get an entry");
	if (verbose) {
	    if (filename)
		if (fprintf(stderr, "%s:\t", filename) < 9)
		    putc('\t', stderr);
	    ss = getname();
	    fputeucs(ss, stderr);
	    putc('\n', stderr);
	    free(ss);
	}
	/*
	 * And do the job.
	 */
	process(iop);
	/*
	 * Close the file.
	 */
	if (iop != stdin)
	    fclose(iop);
    } while (v++, --c > 0);
    if (ferror(stdout))
	error("output write error");
    fclose(defp);
    return status;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	readkeyword - read additional keyword file
 ---------------------------------------------------------------------*/
static char_t *readkeyword(filename)
    char *filename;
{
    register int c;
    register int i;
    int size;
    FILE *fp;
    char_t *buf;

    if ((buf = (char_t *)
	    malloc(sizeof (char_t) * (size = BUFSIZ))) == NULL)
	error("out of memory");
    if (strcmp(filename, "-") == 0)
	fp = stdin;
    else
	if ((fp = fopen(filename, "r")) == NULL) {
	    perror(filename);
	    exit(1);
	}
    for (i = 0; (c = geteucc(fp)) != EOF; ) {
	if (i >= size - 1)
	    if ((buf = (char_t *) realloc(buf,
		    sizeof (char_t) * (size += BUFSIZ))) == NULL)
		error("out of memory");
	buf[i++] = c;
    }
    buf[i++] = '\0';
    if ((buf = (char_t *) realloc(buf, sizeof (char_t) * i)) == NULL)
	error("out of memory");
    if (fp != stdin)
	fclose(fp);
    return buf;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    NAME
	error - print error message on stderr and die
    SYNOPSIS
	error(format, arg1, arg2, ...)
	    char *format;
 ---------------------------------------------------------------------*/
#include <varargs.h>

void error(va_alist)
    va_dcl
{
    va_list args;

    va_start(args);
    fprintf(stderr, "%s: ", progname);
    vfprintf(stderr, va_arg(args, char *), args);
    putc('\n', stderr);
    va_end(args);
    exit(1);
}

/**********************************************************************
 *                                                                    *
 *  Default Language Routines                                         *
 *                                                                    *
 **********************************************************************/
struct langtable {
    union regexp *fname;
    union regexp *lname;
    long offset;
};
static struct langtable *langtable;
static int numlang;

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	makelangtable - make language name table
 ---------------------------------------------------------------------*/
static void makelangtable(fp)
    FILE *fp;
{
    register int i = 0;
    union regexp *lname, *fname;
    long offset;

    if ((langtable = (struct langtable *)
	    malloc(sizeof (struct langtable) * (numlang = TABLEALLOC)))
	    == NULL)
	error("out of memory");
    for (i = 0; (offset = getentry(fp, -1L)) >= 0; ) {
	lname = compile("nm", 0);
	fname = compile("fn", 0);
	if (lname || fname) {
	    if (i >= numlang)
		if ((langtable = (struct langtable *)
			realloc(langtable, sizeof (struct langtable)
			* (numlang += TABLEALLOC))) == NULL)
		    error("out of memory");
	    langtable[i].lname = lname;
	    langtable[i].fname = fname;
	    langtable[i].offset = offset;
	    i++;
	}
    }
    if ((langtable = (struct langtable *) realloc(langtable,
	    sizeof (struct langtable) * (numlang = i))) == NULL)
	error("out of memory");
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	deflang - default language
    DESCRIPTION
	Decides the language type by langage name or file name.
 ---------------------------------------------------------------------*/
static long deflang(name, type)
    char *name;
    enum nametype type;
{
    register int i;
    register char_t *p;
#if defined EUC
    char *s;
#endif
    char_t buf[BUFSIZ];

#if defined EUC
    for (s = name, p = buf; *s; p++)
	*p = pcode(&s);
/*
    *p++ = '\n';
*/
    *p = '\0';
#else
    strcpy(buf, name);
#endif
    for (i = 0; i < numlang; i++)
	if ((p = regmatch(buf, buf, type == FNAME ? langtable[i].fname
		: langtable[i].lname)) && *p == '\0')
	    return langtable[i].offset;
    return -1L;
}

/**********************************************************************
 *                                                                    *
 *  Routines which Actually Do the Job                                *
 *                                                                    *
 **********************************************************************/
static int (*cmpfunc)();

static int sortfunc();
static void putstr();
static bool iskeyword();
static char_t **makekwdtable();

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	process - body of "cmt"
 ---------------------------------------------------------------------*/
static void process(iop)
    FILE *iop;
{
    register char_t *s, *ss;
    char_t *end;
    char_t line[BUFSIZ];	/* line buffer */
    bool loop, inpp;
    bool allid;			/* "all identifiers" flag */
    char_t *kwd = NULL;
    char_t **kwdtable = NULL;
    int numkwd = 0;
    char_t *pkwd = NULL;
    char_t **pkwdtable = NULL;
    int numpkwd = 0;
    /*
     * Regular expressions:
     */
    union regexp *r_id;		/* identifier */
    union regexp *r_nc;		/* numeric constant */
    union regexp *r_cb;		/* string introducing a comment */
    union regexp *r_ce;		/* string ending a comment */
    union regexp *r_ab;		/* string introducing a comment */
    union regexp *r_ae;		/* string ending a comment */
    union regexp *r_sb;		/* beg delimiter for string constant */
    union regexp *r_se;		/* end delimiter for string constant */
    union regexp *r_sn;		/* ignored in string constant */
    union regexp *r_hb;		/* beg delimiter for char constant */
    union regexp *r_he;		/* end delimiter for char constant */
    union regexp *r_hn;		/* ignored in char constant */
    union regexp *r_pb;		/* beg delimiter for a preproc line */
    union regexp *r_pe;		/* end delimiter for a preproc line */
    union regexp *r_pn;		/* ignored in a preproc line */

    /*
     * Check "one case" flag.
     */
#if defined EUC
    cmpfunc = onecase
	    || (onecase = getflag("oc")) ? euccasecmp : euccmp;
#else
    cmpfunc = onecase
	    || (onecase = getflag("oc")) ? strcasecmp : strcmp;
#endif
    /*
     * Check "all identifiers" flag, which makes all identifiers
     * matched boldfaced.
     */
    allid = getflag("al");
    /*
     * If the flag is not set, reads keywords into buffers.
     */
    if (keyword && !allid) {
	/*
	 * Read keywords into buffer, and create table.
	 */
	if (kwd = getstr("kw"))
	    kwdtable = makekwdtable(NULL, kwd, &numkwd);
	if (addkwd)
	    kwdtable = makekwdtable(kwdtable, addkwd, &numkwd);
	qsort(kwdtable, numkwd, sizeof (char_t **), sortfunc);
	/*
	 * Read preprocessor-line keywords into buffer, and create
	 * table.
	 */
	if (pkwd = getstr("pw"))
	    pkwdtable = makekwdtable(NULL, pkwd, &numpkwd);
	qsort(pkwdtable, numpkwd, sizeof (char_t **), sortfunc);
	if (numkwd == 0 && numpkwd == 0)
	    keyword = 0;
    }
    /*
     * Compile regular expressions.
     */
    r_id = compile("id", onecase);
    r_nc = compile("nc", onecase);
    r_cb = compile("cb", onecase);
    r_ce = compile("ce", onecase);
    r_ab = compile("ab", onecase);
    r_ae = compile("ae", onecase);
    r_sb = compile("sb", onecase);
    r_se = compile("se", onecase);
    r_sn = compile("sn", onecase);
    r_hb = compile("hb", onecase);
    r_he = compile("he", onecase);
    r_hn = compile("hn", onecase);
    r_pb = compile("pb", onecase);
    r_pe = compile("pe", onecase);
    r_pn = compile("pn", onecase);
    /*
     * Read and process!!
     */
    s = line;
    *s = '\0';
    inpp = 0;
    while (*s || (s = fgeteucs(line, numelemof (line), iop))) {
	/*
	 * See if it is the beginning of comment.
	 */
	if (r_cb && (ss = regmatch(line, s, r_cb))) {
	    for (loop = 1; ; ) {
		for (; *ss; ss++)
		    if (r_ce && (end = regmatch(line, ss, r_ce))) {
			ss = end;
			loop = 0;
			break;
		    }
		putstr(s, ss, UNDERLINE);
		if (!loop)
		    break;
		if (fgeteucs(line, numelemof (line), iop) == NULL)
		    goto eof;
		s = ss = line;
	    }
	    s = ss;
	/*
	 * See if it is the beginning of alternative comment.
	 */
	} else if (r_ab && (ss = regmatch(line, s, r_ab))) {
	    for (loop = 1; ; ) {
		for (; *ss; ss++)
		    if (r_ae && (end = regmatch(line, ss, r_ae))) {
			ss = end;
			loop = 0;
			break;
		    }
		putstr(s, ss, UNDERLINE);
		if (!loop)
		    break;
		if (fgeteucs(ss = line, numelemof (line), iop) == NULL)
		    goto eof;
		s = ss = line;
	    }
	    s = ss;
	/*
	 * See if it is the beginning of string constant.
	 */
	} else if (r_sb && (ss = regmatch(line, s, r_sb))) {
	    for (loop = 1; ; ) {
		while (*ss)
		    if (r_sn && (end = regmatch(line, ss, r_sn)))
			ss = end;
		    else if (r_se && (end = regmatch(line, ss, r_se))) {
			ss = end;
			loop = 0;
			break;
		    } else
			ss++;
		putstr(s, ss, 0);
		if (!loop)
		    break;
		if (fgeteucs(line, numelemof (line), iop) == NULL)
		    goto eof;
		s = ss = line;
	    }
	    s = ss;
	/*
	 * See if it is the beginning of character constant.
	 */
	} else if (r_hb && (ss = regmatch(line, s, r_hb))) {
	    for (loop = 1; ; ) {
		while (*ss)
		    if (r_hn && (end = regmatch(line, ss, r_hn)))
			ss = end;
		    else if (r_he && (end = regmatch(line, ss, r_he))) {
			ss = end;
			loop = 0;
			break;
		    } else
			ss++;
		putstr(s, ss, 0);
		if (!loop)
		    break;
		if (fgeteucs(line, numelemof (line), iop) == NULL)
		    goto eof;
		s = ss = line;
	    }
	    s = ss;
	/*
	 * See if it is numeric constant.
	 */
	} else if (r_nc && (ss = regmatch(line, s, r_nc))) {
	    putstr(s, ss, 0);
	    s = ss;
	/*
	 * And...
	 */
	} else if (inpp) {
	    /*
	     * If in preprocessor line...
	     */
	    if (r_pn && (ss = regmatch(line, s, r_pn))) {
		putstr(s, ss, 0);
		s = ss;
	    /*
	     * See if it is the end of preprocessor line.
	     */
	    } else if (r_pe && (ss = regmatch(line, s, r_pe))) {
		putstr(s, ss, 0);
		s = ss;
		inpp = 0;
	    /*
	     * See if it is an identifier.
	     */
	    } else if (r_id && (ss = regmatch(line, s, r_id))) {
		putstr(s, ss, keyword && (allid || iskeyword(pkwdtable,
			numpkwd, s, ss)) ? BOLDFACE : 0);
		s = ss;
	    /*
	     * And the rest.
	     */
	    } else
#if defined EUC
		*s & 0x80 ? puteucchar(*s++) : putchar(*s++);
#else
		putchar(*s++);
#endif
	} else {
	    /*
	     * If not in preprocessor line...
	     *
	     * See if it is the beginning of preprocessor line.
	     */
	    if (r_pb && (ss = regmatch(line, s, r_pb))) {
		putstr(s, ss, 0);
		s = ss;
		inpp = 1;
	    /*
	     * See if it is an identifier.
	     */
	    } else if (r_id && (ss = regmatch(line, s, r_id))) {
		putstr(s, ss, keyword && (allid || iskeyword(kwdtable,
			numkwd, s, ss)) ? BOLDFACE : 0);
		s = ss;
	    /*
	     * And the rest.
	     */
	    } else
#if defined EUC
		*s & 0x80 ? puteucchar(*s++) : putchar(*s++);
#else
		putchar(*s++);
#endif
	}
    }
eof:
    /*
     * Free regular expressions.
     */
    regfree(r_id);
    regfree(r_nc);
    regfree(r_cb);
    regfree(r_ce);
    regfree(r_ab);
    regfree(r_ae);
    regfree(r_sb);
    regfree(r_se);
    regfree(r_sn);
    regfree(r_hb);
    regfree(r_he);
    regfree(r_hn);
    regfree(r_pb);
    regfree(r_pe);
    regfree(r_pn);
    if (kwd)
	free(kwd);
    if (kwdtable)
	free(kwdtable);
    if (pkwd)
	free(pkwd);
    if (pkwdtable)
	free(pkwdtable);
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	sortfunc
 ---------------------------------------------------------------------*/
static int sortfunc(p1, p2)
    char_t **p1, **p2;
{
    return (*cmpfunc)(*p1, *p2);
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	makekwdtable
    DESCRIPTION
	Stores pointer to each keyword into newly allocated or
	re-allocated table area.  Returns pointer to the area.
 ---------------------------------------------------------------------*/
static char_t **makekwdtable(table, str, size)
    char_t **table;
    char_t *str;
    int *size;
{
    register int i = 0;
    register char_t *s;

    if (table) {
	i = *size;
	table = (char_t **) realloc(table, sizeof (char_t *)
		* (i + (str ? (streuclen(str) + 1) / 2: 0)));
    } else
	table = (char_t **) malloc(sizeof (char_t *)
		* (str ? (streuclen(str) + 1) / 2: 0));
    if (table == NULL)
	error("out of memory");
    for (s = str; *s; ) {
	while (*s < 0x80 && isspace(*s))
	    *s++ = '\0';
	if (*s)
	    table[i++] = s;
	while (*s && (*s >= 0x80 || !isspace(*s)))
	    s++;
    }
    if ((table = (char_t **)
	    realloc(table, sizeof (char_t *) * i)) == NULL)
	error("out of memory");
    *size = i;
    return table;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	putstr
 ---------------------------------------------------------------------*/
static void putstr(s, end, type)
    register char_t *s, *end;	/* beg & end of string to write */
    register unsigned type;
{
#if defined EUC
    register int w;
    int width;
#endif

    for (; *s && s < end; s++) {
#if defined EUC
	if ((width = euccol(*s)) > 1) {
	    /*
	     * For characters with two or more display width, the
	     * character is printed first to prevent confusing commands
	     * like fold.
	     */
	    puteucchar(*s);
	    if (type & BOLDFACE) {
		for (w = width; w; --w)
		    putchar('\b');
		puteucchar(*s);
	    }
	    if (type & UNDERLINE) {
		for (w = width; w; --w)
		    putchar('\b');
		for (w = width; w; --w)
		    putchar('_');
	    }
	    continue;
	}
#endif
	if (
#if defined EUC
		*s & 0x80 ||
#endif
		isgraph(*s)) {
	    if (type & UNDERLINE) {
		putchar('_');
		putchar('\b');
	    }
	    if (type & BOLDFACE) {
#if defined EUC
		*s & 0x80 ? puteucchar(*s) : putchar(*s);
#else
		putchar(*s);
#endif
		putchar('\b');
	    }
	}
#if defined EUC
	*s & 0x80 ? puteucchar(*s) : putchar(*s);
#else
	putchar(*s);
#endif
    }
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	iskeyword - check and see if the next word is a keyword
 ---------------------------------------------------------------------*/
static bool iskeyword(table, num, word, end)
    char_t **table;
    int num;
    register char_t *word;
    char_t *end;
{
    register int i, high, low;
    int c, stat;

    c = *end;
    *end = '\0';
    low = 0;
    high = num - 1;
    while (low <= high) {
	i = (low + high) / 2;
	if ((stat = (*cmpfunc)(word, table[i])) < 0)
	    high = i - 1;
	else if (stat > 0)
	    low = i + 1;
	else {
	    *end = c;
	    return 1;
	}
    }
    *end = c;
    return 0;
}

/**********************************************************************
 *                                                                    *
 *  Regular Expression Compilation                                    *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	compile
 ---------------------------------------------------------------------*/
static union regexp *compile(name, oc)
    char *name;
    bool oc;
{
    register char_t *str;
    union regexp *r;

    if (str = getstr(name)) {
	if (r = regcomp(str, oc)) {
	    free(str);
	    return r;
	}
	fprintf(stderr, "%s: %s: syntax error: \"%s=",
		progname, deffile, name);
	fputeucs(str, stderr);
	fputs("\"\n", stderr);
	exit(1);
    }
    return NULL;
}
