/**** << Text Formating & Printing Utilities >> ****
 *
 *  defs.c
 *                                                  Jul 02 1983
 *                                      mod:        Sep 02 1993
 ************************************************** tonooka ***********/
/*
 *	Copyright (c) 1994 Yasuhiro Tonooka (tonooka@msi.co.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#if !defined lint
static char sccsid[] = "@(#)defs.c 2.2 (Y.Tonooka) 3/28/94";
#endif

#include <stdio.h>
#include <ctype.h>
#include "euc.h"
#include "cmt.h"

typedef unsigned char u_char;

/*
 * Routines for dealing with the language definition data base.
 *
 * See "cmtdefs" file for file format.
 */

static char_t *entry;		/* buffer */
static int entsize = 0;		/* size of buffer */

static bool isempty();
static bool namematch();
static char_t *skip();

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    NAME
	getentry
    DESCRIPTION
	Reads non-empty entry into allocated buffer.
	Starts reading at "offset" if "offset" is not negative,
	if not, at the current position.
	Returns the offset of the beginning of an entry relative to
	the beginning of the file associated with "fp".
	Escaped newlines are eliminated.
	-1 is returned for improper position or at the EOF.
    ARGUMENTS
	(i)   FILE *	: fp		: file pointer
	(i)   long	: offset	: position
    RETURNS
	      long			: position
    NOTES
	"error()" is called at ``out of memory'' error.
 ---------------------------------------------------------------------*/
long getentry(fp, offset)
    FILE *fp;
    long offset;
{
    register int c;
    register int i;

    if (offset >= 0)
	if (fseek(fp, offset, 0) < 0)
	    return -1L;
    if (entsize == 0)
	if ((entry = (char_t *)
		malloc(sizeof (char_t) * (entsize = BUFSIZ))) == NULL)
	    error("out of memory");
    do {
	offset = ftell(fp);
	for (i = 0; ; ) {
	    if ((c = geteucc(fp)) == EOF)
		return -1L;
	    if (c == '\n') {
		if (i > 0 && entry[i - 1] == '\\') {
		    --i;
		    continue;
		}
		break;
	    }
	    if (i >= entsize - 1)
		if ((entry = (char_t *) realloc(entry,
			sizeof (char_t) * (entsize += BUFSIZ))) == NULL)
		    error("out of memory");
	    entry[i++] = c;
	}
	entry[i] = '\0';
    } while (isempty(entry));
    return offset;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	isempty - see if the entry is empty
 ---------------------------------------------------------------------*/
static bool isempty(str)
    char_t *str;
{
    if (*str == '#')
	return 1;
    for (; *str; str++)
	if (*str >= 0x80 || isspace(*str))
	    return 0;
    return 1;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    NAME
	getname - get the name field
    DESCRIPTION
	Returns the name field (the first field).
 ---------------------------------------------------------------------*/
char_t *getname()
{
    register char_t *p;
    register int i;
    char_t *area;
    int size;

    if ((area = (char_t *)
	    malloc(sizeof (char_t) * (size = BUFSIZ))) == NULL)
	error("out of memory");
    for (i = 0, p = entry; *p && *p != ':'; i++) {
	if (i >= size - 1)
	    if ((area = (char_t *) realloc(area, sizeof (char_t)
		    * (size += BUFSIZ))) == NULL)
		error("out of memory");
	area[i] = *p++;
    }
    area[i] = '\0';
    if ((area = (char_t *)
	    realloc(area, sizeof (char_t) * (i + 1))) == NULL)
	error("out of memory");
    return area;
}

#if 0				/* NOT USED */
/*---------------------------------------------------------------------
    NAME
	getnum - return the numeric id
    DESCRIPTION
	Numeric options look like
	    li#80
	i.e., the option string is separated from the numeric value by
	a `#' character.  If the option is not found it returns -1.
	Note that it handles octal numbers beginning with 0.
 ---------------------------------------------------------------------*/
int getnum(id)
    char *id;
{
    register int i, base;
    register char_t *p;

    p = entry;
    do {
	if (!*(p = skip(p)))
	    return -1;
    } while (*p++ != ((u_char *) id)[0] || !*p
	    || *p++ != ((u_char *) id)[1] || *p != '#');
    p++;
    if (*p == '0')
	base = 8;
    else
	base = 10;
    for (i = 0; *p < 0x80 && isdigit(*p); )
	i *= base, i += *p++ - '0';
    return i;
}
#endif

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    NAME
	getflag - handle a flag option
    DESCRIPTION
	Flag options are given "naked", i.e. followed by a : or the end
	of the buffer.  Return 1 if we find the option, or 0 if it is
	not given.
 ---------------------------------------------------------------------*/
bool getflag(id)
    char *id;
{
    register char_t *p;

    for (p = entry; *(p = skip(p)); )
	if (*p++ == ((u_char *) id)[0] && *p
		&& *p++ == ((u_char *) id)[1])
	    if (!*p || *p == ':')
		return 1;
    return 0;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    NAME
	getstr - get a string valued option
    DESCRIPTION
	These are given as
	    cb=/\*
	Much decoding is done on the strings, and the strings are
	placed on newly allocated area.
 ---------------------------------------------------------------------*/
char_t *getstr(id)
    char *id;
{
    register char_t *p;
    register int i;
    char_t *area;
    int size;

    if ((area = (char_t *)
	    malloc(sizeof (char_t) * (size = BUFSIZ))) == NULL)
	error("out of memory");
    for (p = entry; *(p = skip(p)); )
	if (*p++ == ((u_char *) id)[0] && *p
		&& *p++ == ((u_char *) id)[1] && *p == '=') {
	    for (i = 0, p++; *p && *p != ':'; i++) {
		if (i >= size - 1)
		    if ((area = (char_t *) realloc(area, sizeof (char_t)
			    * (size += BUFSIZ))) == NULL)
			error("out of memory");
		area[i] = *p++;
	    }
	    area[i] = '\0';
	    if ((area = (char_t *)
		    realloc(area, sizeof (char_t) * (i + 1))) == NULL)
		error("out of memory");
	    return area;
	}
    return NULL;
}

/*-==(spec:\\:":":\(:\):\[:\]::::cmt)==--------------------------------
    TYPE
	local function
    NAME
	skip - skip to the next field
 ---------------------------------------------------------------------*/
static char_t *skip(p)
    register char_t *p;
{
    while (*p && *p != ':')
	p++;
    if (*p == ':')
	p++;
    return p;
}
