/**** << Text Formating & Printing Utilities >> ****
 *
 *  euc.c
 *                                                  Aug 16 1993
 *                                      mod:        Aug 19 1993
 ************************************************** tonooka ***********/
/*
 *	Copyright (c) 1994 Yasuhiro Tonooka (tonooka@msi.co.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#if !defined lint
static char *sccsid = "@(#)euc.c 2.2 (Y.Tonooka) 3/28/94";
#endif

#include <stdio.h>
#include "euc.h"

#if defined EUC

typedef unsigned char u_char;

#define SS2	0x8e
#define SS3	0x8f

/**********************************************************************
 *                                                                    *
 *  Compatibility                                                     *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	pcode
 ---------------------------------------------------------------------*/
int pcode(s)
    register char **s;
{
    register int c0, c1;

    if ((c0 = *(u_char *) (*s)++) & 0x80) {
	if ((c1 = *(u_char *) (*s)++) & 0x80)
	    switch (c0) {
	    case SS2:
		return c1;
	    case SS3:
		if (*(u_char *) (*s) & 0x80)
		    return c1 << 8 | *(u_char *) (*s)++ & 0x7f;
		break;
	    default:
		return c0 << 8 | c1;
	    }
	--(*s);
	c0 = ' ';
    }
    return c0;
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	geteucc
 ---------------------------------------------------------------------*/
int geteucc(fp)
    register FILE *fp;
{
    register int c, cc;

    if ((c = getc(fp)) == EOF || c < 0x80)
	return c;
    if ((cc = getc(fp)) == EOF)
	return EOF;
    switch (c) {
    case SS2:
	return cc;
    case SS3:
	if ((c = getc(fp)) == EOF)
	    return EOF;
	return cc << 8 | c & 0x7f;
    default:
	return c << 8 | cc;
    }
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	puteucc
 ---------------------------------------------------------------------*/
int puteucc(c, fp)
    register int c;
    FILE *fp;
{
    if (c < 0x80)		/* ASCII */
	return putc(c, fp);
    if (c < 0x100) {		/* SS2 */
	putc(SS2, fp);
	return putc(c, fp);
    }
    if (c & 0x80) {		/* SS1 */
	putc(c >> 8, fp);
	return putc(c & 0xff, fp);
    }
    putc(SS3, fp);		/* SS3 */
    putc(c >> 8, fp);
    return putc(c & 0xff | 0x80, fp);
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	euccol - gives displaywidth of an EUC character
 ---------------------------------------------------------------------*/
int euccol(c)
    register int c;
{
    return c < 0x100 ? 1 : 2;
}

/**********************************************************************
 *                                                                    *
 *  Other Routines                                                    *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	fgeteucs
 ---------------------------------------------------------------------*/
char_t *fgeteucs(str, n, fp)
    char_t *str;
    register int n;
    register FILE *fp;
{
    register c;
    register char_t *s;

    s = str;
    while (--n > 0 && (c = geteucc(fp)) != EOF) {
	*s++ = c;
	if (c == '\n')
	    break;
    }
    if (c == EOF && s == str)
	return NULL;
    *s = '\0';
    return str;
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	fputeucs
 ---------------------------------------------------------------------*/
int fputeucs(s, fp)
    register char_t *s;
    register FILE *fp;
{
    register int c, r;

    while (c = *s++)
	r = puteucc(c, fp);
    return(r);
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	streuclen
 ---------------------------------------------------------------------*/
int streuclen(s)
    register char_t *s;
{
    register int i = 0;

    while (*s++)
	i++;
    return i;
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	euccmp
 ---------------------------------------------------------------------*/
int euccmp(s1, s2)
    register char_t *s1, *s2;
{
    for (; *s1 == *s2; s1++, s2++)
	if (*s1 == '\0')
	    return 0;
    return *s1 - *s2;
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	euccasecmp
 ---------------------------------------------------------------------*/
int euccasecmp(s1, s2)
    register char_t *s1, *s2;
{
    register int c;

    for (; (c = (*s1 < 0x80 ? toupper(*s1) : *s1)
	    - (*s2 < 0x80 ? toupper(*s2) : *s2)) == 0; s1++, s2++)
	if (*s1 == '\0')
	    return 0;
    return c;
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	stow - single-byte-char string to processing code string
 ---------------------------------------------------------------------*/
char_t *stow(s)
    char *s;
{
    register char_t *w;
    char_t *euc;

    if ((euc = (char_t *) malloc(sizeof (char_t) * (strlen(s) + 1)))
	    == NULL)
	return NULL;
    for (w = euc; *s; )
	*w++ = *(u_char *) s < 0x80 ? *s++ : pcode(&s);
    *w++ = '\0';
    return (char_t *) realloc(euc, sizeof (char_t) * (w - euc));
}

/*-==(spec:\\:":":\(:\):\[:\]::::euc)==--------------------------------
    NAME
	eucdup
 ---------------------------------------------------------------------*/
char_t *eucdup(s)
    register char_t *s;
{
    register char_t *d;
    char_t *dup;

    if (dup = (char_t *) malloc(sizeof (char_t) * (streuclen(s) + 1)))
	for (d = dup; ; )
	    if (!(*d++ = *s++))
		break;
    return dup;
}

#endif
