/**** << Text Formating & Printing Utilities >> ****
 *
 *  prn.c
 *                                                  Aug 23 1993
 *                                      mod:        May 11 1994
 ************************************************** tonooka ***********/
/*
 *	Copyright (c) 1994 Yasuhiro Tonooka (tonooka@msi.co.jp)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#if !defined lint
static char sccsid[] = "@(#)prn.c 2.3 (Y.Tonooka) 5/26/94";
#endif

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "euc.h"

#define KCCPATH		"/usr/local/bin/kcc"
#if defined (__bsdi__)
#define LPRPATH		"/usr/bin/lpr"
#else
#define LPRPATH		"/usr/ucb/lpr"
#endif

#if defined DEBUG

#define CMTPATH		"./cmt"
#define PSPATH		"./psconv"

#else

#define CMTPATH		"/usr/local/bin/cmt"
#define PSPATH		"/usr/local/bin/psconv"

#endif

#define MAXARGS		64

typedef int bool;

#define numelemof(arr)   (sizeof (arr) / sizeof (*(arr)))

extern long strtol();
extern double strtod();
extern time_t time();

static char *progname;

static bool dokcc = 1;
static bool docmt = 0;
static bool dops = 1;
static bool dolpr = 1;
static bool langopt = 0;
static bool bundle = 0;

static char *cmtargs[MAXARGS] = { CMTPATH, };
static char *psargs[MAXARGS] = { PSPATH, };
static char *lprargs[MAXARGS] = { LPRPATH, };

static int numcmtargs = 1;
static int numpsargs = 1;
static int numlprargs = 1;

static char *kccargs[] = { KCCPATH, "-e", (char *) NULL};

static bool donumber = 0;
static int tabstop = 8;

static int width = 80;
static int length = 66;

static bool doheader = 1;
static char *header = NULL;

static char *rangestr = NULL;

static char *filename;		/* name of the file being processed */
static char *modtime;		/* the last modification time */

static void error();
static char *basename();
static bool formatopt();
static bool rangeopt();

static int infilter();
static int outfilter();
static int outlpr();

static void process();

static char *getrange();

static char *strtime();

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    NAME
	main
 ---------------------------------------------------------------------*/
int main(c, v)
    int c;
    register char *v[];
{
    register int i;
    register char *s;
    int ind, outd;
    FILE *inp, *outp;
    int ret;
    int child[4];
    int numchild = 0;
    int pid;
    int dev, ino = -1;
    struct stat sbuf;
    int status = 0;

    for (i = 0; i < numelemof (child); i++)
	child[i] = 0;
    progname = *v++;
    if (strcmp(basename(progname), "cprn") == 0)
	docmt = 1;
    /*
     * Process command line options.
     */
    for (; --c; v++) {
	/*
	 * With -, input is taken from stdin like cat(1).
	 */
	if ((*v)[0] != '-' || (*v)[1] == '\0')
	    break;
	/*
	 * Format specificaion:
	 */
	if ((*v)[1] == 'V') {
	    if ((*v)[2] == '\0') {
		width = 80;
		length = 66;
		psargs[numpsargs++] = "-V";
	    } else
		if (strcmp(*v + 2, "w") == 0) {
		    width = 132;
		    length = 100;
		    psargs[numpsargs++] = "-Vw";
		} else if (strcmp(*v + 2, "2") == 0) {
		    width = 80;
		    length = 132;
		    psargs[numpsargs++] = "-V2";
		    psargs[numpsargs++] = "-n";
		} else if (strcmp(*v + 2, "3") == 0) {
		    width = 80;
		    length = 200;
		    psargs[numpsargs++] = "-V3";
		    psargs[numpsargs++] = "-n";
		    psargs[numpsargs++] = "-m6";
		} else
		    if (formatopt(*v))
			error("%s: bad format", *v);
	    continue;
	}
	if ((*v)[1] == 'H') {
	    if ((*v)[2] == '\0') {
		width = 132;
		length = 66;
		psargs[numpsargs++] = "-H";
	    } else
		if (strcmp(*v + 2, "2") == 0) {
		    width = 80;
		    length = 66;
		    psargs[numpsargs++] = "-H2";
		} else if (strcmp(*v + 2, "3") == 0) {
		    width = 80;
		    length = 100;
		    psargs[numpsargs++] = "-H3";
		    psargs[numpsargs++] = "-n";
		    psargs[numpsargs++] = "-m6";
		} else if (strcmp(*v + 2, "4") == 0) {
		    width = 80;
		    length = 132;
		    psargs[numpsargs++] = "-H4";
		    psargs[numpsargs++] = "-n";
		    psargs[numpsargs++] = "-m6";
		} else
		    if (formatopt(*v))
			error("%s: bad format", *v);
	    continue;
	}
	/*
	 * Font specification:
	 */
	if ((*v)[1] == 'f') {
	    psargs[numpsargs++] = *v;
	    continue;
	}
	/*
	 * lpr options:  -P & -#
	 */
	if ((*v)[1] == 'P' || (*v)[1] == '#') {
	    lprargs[numlprargs++] = *v;
	    continue;
	}
	/*
	 * cmt option:  -l
	 */
	if ((*v)[1] == 'l') {
	    langopt = 1;
	    cmtargs[numcmtargs++] = *v;
	    continue;
	}
	/*
	 * Specify the language definition file.
	 */
	if (strcmp(*v, "-d") == 0) {
	    if (--c == 0)
		error("%s option must have an argument", *v);
	    cmtargs[numcmtargs++] = *v++;
	    cmtargs[numcmtargs++] = *v;
	    continue;
	}
	/*
	 * Specify the header string.
	 */
	if (strcmp(*v, "-h") == 0) {
	    if (--c == 0)
		error("%s option must have an argument", *v);
	    v++;
	    header = *v;
	    continue;
	}
	/*
	 * Specify ranges to print.
	 */
	if ((*v)[1] == 'o') {
	    if (rangeopt(*v))
		error("%s: bad range", *v);
	    rangestr = *v + 2;
	    continue;
	}
	/*
	 * Tabstops are assumed to be at every eighth character
	 * posision unless they are specified with option -`num'.
	 */
	if (isdigit(*(u_char *) (*v + 1))) {
	    if ((tabstop = atoi(*v + 1)) < 1)
		error("%s: bad tabstop", *v);
	    continue;
	}
	/*
	 * One letter options:
	 */
	for (s = *v + 1; *s; s++)
	    switch (*s) {
	    case 'a':
		/*
		 * ASCII output.
		 */
		dops = dolpr = 0;
		break;
	    case 'B':
		/*
		 * Bundle
		 */
		bundle = 1;
		break;
	    case 'b':
		/*
		 * "psconv" option:  `-fB:I:BIU:U'.
		 */
		psargs[numpsargs++] = "-fB:I:BIU:U";
		break;
	    case 'c':
		/*
		 * Go thru "cmt".
		 */
		docmt = 1;
		break;
	    case 'i':
		/*
		 * "cmt" option:  case insensitive.
		 */
		cmtargs[numcmtargs++] = "-i";
		break;
	    case 'j':
		/*
		 * "psconv" option:  square kanji.
		 */
		psargs[numpsargs++] = "-j";
		break;
	    case 'k':
		/*
		 * "cmt" option:  do not emphasize keywords.
		 */
		cmtargs[numcmtargs++] = "-k";
		break;
	    case 'n':
		/*
		 * Line numbering.
		 */
		donumber = 1;
		break;
	    case 'p':
		/*
		 * PostScript output.
		 */
		dolpr = 0;
		break;
	    case 's':
		/*
		 * "psconv" option:  Draw separator(s).
		 */
		psargs[numpsargs++] = "-s";
		break;
	    case 't':
		/*
		 * "pr" option:  No header or trailer.
		 */
		doheader = 0;
		break;
	    case 'u':
		/*
		 * Unfiltered:  Do not "kcc".
		 */
		dokcc = 0;
		break;
	    default:
		s[1] = '\0';
		error("%c: bad option", *s);
	    }
    }
    psargs[numpsargs] = (char *) NULL;
    lprargs[numlprargs] = (char *) NULL;
    if (langopt)
	cmtargs[numcmtargs] = (char *) NULL;
    if (!dolpr && dops && c > 1)
	bundle = 1;
    if (doheader && length < 11)
	error("length must be more than 10");
    /*
     * Get some info on stdout.
     */
    if (!dolpr)
	if (fstat(fileno(stdout), &sbuf) == 0) {
	    sbuf.st_mode &= S_IFMT;
	    if (sbuf.st_mode != S_IFCHR && sbuf.st_mode != S_IFBLK) {
		dev = sbuf.st_dev;
		ino = sbuf.st_ino;
	    }
	}
    /*
     * If bundle, plumb around output and open stdio.
     */
    if (bundle) {
	outd = dolpr ? outlpr((int *) NULL) : 1;
	if (dops)
	    outd = outfilter(outd, psargs, (int *) NULL);
	if ((outp = fdopen(outd == 1 ? dup(outd) : outd, "w"))
		== NULL)
	    error("out of memory");
    }
    /*
     * Main loop.
     */
    do {
	if (c == 0 || strcmp(*v, "-") == 0) {
	    /*
	     * Stdin:
	     */
	    ind = 0;
	    filename = "";
	    cmtargs[numcmtargs] = (char *) NULL;
	} else {
	    /*
	     * Open a file.
	     */
	    if ((ind = open(*v, O_RDONLY, 0)) < 0) {
		perror(*v);
		status = 1;
		continue;
	    }
	    filename = *v;
	    if (!langopt) {
		cmtargs[numcmtargs] = "-L";
		cmtargs[numcmtargs + 1] = *v;
		cmtargs[numcmtargs + 2] = (char *) NULL;
	    }
	}
	/*
	 * Get some info on input file, and see if it is a
	 * directory. 
	 */
	if (fstat(ind, &sbuf) < 0) {
	    perror(filename);
	    exit(1);
	}
	if ((sbuf.st_mode & S_IFMT) == S_IFDIR) {
	    fprintf(stderr,
		    "%s: read error on %s: is a directory\n",
		    progname, c ? *v : "standard input");
	    if (ind != 0)
		close(ind);
	    status = 1;
	    continue;
	}
	/*
	 * Get the time the file was last modified, or current time for
	 * stdin.
	 */
	modtime = strtime(ind == 0 ? time(NULL) : sbuf.st_mtime);
	/*
	 * Compare input with output, and see if they are
	 * identical.
	 */
	if (!dolpr)
	    if ((sbuf.st_mode & S_IFMT) == S_IFREG &&
		    sbuf.st_dev == dev && sbuf.st_ino == ino) {
		fprintf(stderr, "%s: input %s is output\n",
			progname, c ? *v : "-");
		if (ind != 0)
		    close(ind);
		status = 1;
		continue;
	    }
	/*
	 * Plumb around input and open stdio.
	 */
	if (dokcc)
	    ind = infilter(ind, kccargs, &child[numchild++]);
	if (docmt)
	    ind = infilter(ind, cmtargs, &child[numchild++]);
	if ((inp = fdopen(ind == 0 ? dup(ind) : ind, "r")) == NULL)
	    error("out of memory");
	/*
	 * If !bundle, plumb around output and open stdio.
	 */
	if (!bundle) {
	    outd = dolpr ? outlpr(&child[numchild++]) : 1;
	    if (dops)
		outd = outfilter(outd, psargs, &child[numchild++]);
	    if ((outp = fdopen(outd == 1 ? dup(outd) : outd, "w"))
		    == NULL)
		error("out of memory");
	}
	/*
	 * And do the job!
	 */
	process(inp, outp);
	fclose(inp);
	if (!bundle) {
	    if (ferror(outp)) {
		error("output write error");
		exit(1);
	    }
	    fclose(outp);
	}
	for (; numchild; --numchild)
	    if ((pid = wait(&ret)) > 0) {
		if ((ret & 0xff) != 0 || (ret >> 8 & 0xff) != 0)
		    exit(1);
		for (i = 0; child[i] != pid; i++)
		    ;
		child[i] = 0;
	    }
    } while (v++, --c > 0);
    if (bundle) {
	if (ferror(outp)) {
	    error("output write error");
	    exit(1);
	}
	fclose(outp);
	while (wait(&ret) != -1)
	    ;
    }
    return (status);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	error - print error message on stderr and die
    SYNOPSIS
	error(format, arg1, arg2, ...)
	    char *format;
 ---------------------------------------------------------------------*/
#include <varargs.h>

static void error(va_alist)
    va_dcl
{
    va_list args;

    va_start(args);
    fprintf(stderr, "%s: ", progname);
    vfprintf(stderr, va_arg(args, char *), args);
    putc('\n', stderr);
    va_end(args);
    exit(1);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	basename
 ---------------------------------------------------------------------*/
static char *basename(path)
    char *path;
{
    register char *s;

    for (s = path; *s; )
	if (*s++ == '/')
	    path = s;
    return (path);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	formatopt - handle format option
 ---------------------------------------------------------------------*/
static bool formatopt(str)
    char *str;
{
    char *s;
    int col;
    double stretch = 1.0;
    static char arg[40];

    s = str + 2;
    if (*s == '@') {
	if (s++, *s != '-' && (stretch = strtod(s, &s)) <= 0)
	    return (1);
	width = 0;
    } else
	if ((width = strtol(s, &s, 10)) <= 0)
	    return (1);
    if (*s++ != '-')
	return (1);
    if (*s == '@') {
	if (s++, *s && *s != '-' && (stretch = strtod(s, &s)) <= 0)
	    return (1);
	length = 0;
    } else
	if ((length = strtol(s, &s, 10)) <= 0)
	    return (1);
    if (*s == '-') {
	if ((col = strtol(s + 1, &s, 10)) <= 0)
	    return (1);
    } else
	col = 1;
    if (*s || width == 0 && length == 0)
	return (1);
    if (length == 0)
	length = (str[1] == 'V' ? 66 : 33) * width * col
		/ (80 * stretch);
#if 0
	length = (str[1] == 'V' ? (col == 3 ? 200 : 198)
		: (col == 3 ? 100 : 99)) * width * col
		/ (80 * 3 * stretch);
#endif
    else if (width == 0)
	width = 80 * length
		/ ((str[1] == 'V' ? 66 : 33) * col * stretch);
#if 0
	width = (80 * 3) * length
		/ ((str[1] == 'V' ? (col == 3 ? 200 : 198)
		: (col == 3 ? 100 : 99)) * col * stretch);
#endif
    if (width < 2 || length < 1)
	return (1);
    sprintf(arg, "-%c%d-%d-%d", str[1], width, length, col);
    psargs[numpsargs++] = arg;
    if (col == 2 && str[1] == 'H' || col > 2)
	psargs[numpsargs++] = "-n";
    if (col > 2)
	psargs[numpsargs++] = "-m6";
    return (0);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	rangeopt - see if range option is legal
 ---------------------------------------------------------------------*/
static bool rangeopt(str)
    char *str;
{
    long dum;

    str += 2;
    if (*str == '\0')
	return (1);
    do {
	if ((str = getrange(str, &dum, &dum)) == NULL)
	    return (1);
    } while (*str);
    return (0);
}

/**********************************************************************
 *                                                                    *
 *  Plumbing Routines                                                 *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	infilter - input filter
 ---------------------------------------------------------------------*/
static int infilter(fd, filter, cpid)
    int fd;
    char *filter[];
    int *cpid;
{
    int piped[2];
    int pid;

    if (pipe(piped)) {
	perror("pipe");
	exit(1);
    }
    if ((pid = fork()) < 0) {
	perror("fork");
	exit(1);
    } else if (pid == 0) {
	/*
	 * Child process executes "filter[0]".
	 */
	if (fd != 0) {
	    close(0);
	    dup(fd);
	    close(fd);
	}
	close(1);
	dup(piped[1]);
	close(piped[0]);
	close(piped[1]);
	execv(filter[0], filter);
	perror(filter[0]);
	exit(1);
    }
    if (fd != 0)
	close(fd);
    close(piped[1]);
    if (cpid)
	*cpid = pid;
    return (piped[0]);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	outfilter - output filter
 ---------------------------------------------------------------------*/
static int outfilter(fd, filter, cpid)
    int fd;
    char *filter[];
    int *cpid;
{
    int piped[2];
    int pid;

    if (pipe(piped)) {
	perror("pipe");
	exit(1);
    }
    if ((pid = fork()) < 0) {
	perror("fork");
	exit(1);
    } else if (pid == 0) {
	/*
	 * Child process executes "filter[0]".
	 */
	if (fd != 1) {
	    close(1);
	    dup(fd);
	    close(fd);
	}
	close(0);
	dup(piped[0]);
	close(piped[0]);
	close(piped[1]);
	execv(filter[0], filter);
	perror(filter[0]);
	exit(1);
    }
    if (fd != 1)
	close(fd);
    close(piped[0]);
    if (cpid)
	*cpid = pid;
    return (piped[1]);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	outlpr - open pipe to lpr
 ---------------------------------------------------------------------*/
static int outlpr(cpid)
    int *cpid;
{
    int piped[2];
    int pid;

    if (pipe(piped)) {
	perror("pipe");
	exit(1);
    }
    if ((pid = fork()) < 0) {
	perror("fork");
	exit(1);
    } else if (pid == 0) {
	/*
	 * Child process executes `lpr'.
	 */
	close(0);
	dup(piped[0]);
	close(piped[0]);
	close(piped[1]);
	execv(LPRPATH, lprargs);
	perror(LPRPATH);
	exit(1);
    }
    close(piped[0]);
    if (cpid)
	*cpid = pid;
    return (piped[1]);
}

/**********************************************************************
 *                                                                    *
 *  Processing                                                        *
 *                                                                    *
 **********************************************************************/
static int page, row, col;
static long lineno;		/* line number */
static int left;		/* left margin */

static bool outofrange();
static bool inrange();

static void ln();
static bool cr();
static void head(), tail();
static void puthead();

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	process - does detabbing, line-numbering, folding and handles
		heading
    DESCRIPTION
	"process()" does the job almost equivalent to"

	    expand | [cat -n |] fold [| pr]"
 ---------------------------------------------------------------------*/
static void process(in, out)
    FILE *in, *out;
{
    long from, to;
    char *range;

    lineno = 1;
    page = 0;
    row = -1;
    col = donumber ? -1 : 0;
    if (rangestr)
	range = getrange(rangestr, &from, &to);
    else {
	range = NULL;
	from = 1, to = 0;
    }
    do {
	if (lineno < from)
	    if (outofrange(from, in))
		break;
	if (inrange(to, in, out))
	    break;
    } while (range && (range = getrange(range, &from, &to)));
    if (page > 0 && row >= 0)
	tail(out);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	outofrange
    DESCRIPTION
 ---------------------------------------------------------------------*/
static bool outofrange(from, in)
    long from;
    FILE *in;
{
    register int c;

    while ((c = geteucc(in)) != EOF) {
	if (c == '\n') {
	    lineno++;
	    if (lineno >= from)
		return (0);
	}
    }
    return (1);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	inrange -
    DESCRIPTION
 ---------------------------------------------------------------------*/
static bool inrange(to, in, out)
    long to;
    FILE *in, *out;
{
    register int c;
    int n;

    while ((c = geteucc(in)) != EOF) {
	if (row < 0)
	    head(out);
	if (col < 0)
	    ln(out);
	if (c < 0x80) {
	    switch (c) {
	    case '\n':		/* line feed */
		cr(out);
		col = donumber ? -1 : 0;
		lineno++;
		if (to > 0 && lineno > to)
		    return (0);
		continue;
	    case '\f':		/* form feed */
		tail(out);
		continue;
	    case '\r':		/* carriage return */
		col = left;
		break;
	    case '\b':		/* backspace */
		if (col > left)
		    --col;
		break;
	    case '\t':		/* tab */
		if (col == width)
		    if (cr(out))
			head(out);
		do {
		    putc(' ', out);
		    col++;
		} while ((col - left) % tabstop != 0 && col < width);
		continue;
	    default:		/* other ASCII character */
		if (col == width)
		    if (cr(out))
			head(out);
		col++;
		break;
	    }
	    putc(c, out);
	} else {		/* EUC character */
	    if (col + (n = euccol(c)) > width)
		if (cr(out))
		    head(out);
	    col += n;
	    puteucc(c, out);
	}
    }
    return (1);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	getrange
 ---------------------------------------------------------------------*/
static char *getrange(str, from, to)
    char *str;
    long *from, *to;
{
    char *s;

    s = str;
    *from = (*s == '-') ? 0 : strtol(s, &s, 10);
    if (*s == '-') {
	if ((*to = strtol(s + 1, &s, 10)) < 0)
	    return (NULL);
    } else if (*s == '+') {
	if ((*to = strtol(s + 1, &s, 10)) < 0)
	    return (NULL);
	*to += *from;
    } else
	*to = *from;
    if (s == str)
	return (NULL);
    if (*s)
	if (*s == ',')
	    s++;
	else
	    return (NULL);
    return (s);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	ln - print line number
 ---------------------------------------------------------------------*/
static void ln(out)
    FILE *out;
{
    fprintf(out, "%6ld  ", lineno);
    left = col = 8;
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	cr - carriage return
 ---------------------------------------------------------------------*/
static bool cr(out)
    FILE *out;
{
    putc('\n', out);
    row++;
    col = left = 0;
    if (row == (doheader ? length - 5 : length)) {
	tail(out);
	return (1);
    }
    return (0);
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	head - print header
 ---------------------------------------------------------------------*/
static void head(out)
    FILE *out;
{
    char buf[20];
    page++;
    if (doheader) {
	putc('\n', out);
	putc('\n', out);
	row = 2;
	col = 0;
	puthead(modtime, out);
	puthead("  ", out);
	puthead(header ? header : filename, out);
	sprintf(buf, " Page %d", page);
	puthead(buf, out);
	for (; row < 5; row++)
	    putc('\n', out);
	col = donumber ? -1 : 0;
    } else
	row = 0;
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	puthead - put characters in a header string one by one
 ---------------------------------------------------------------------*/
static void puthead(str, out)
    char *str;
    FILE *out;
{
    int c, n;

    if (row >= 5)
	return;
    for (; *str; ) {
	c = pcode(&str);
	if (col + (n = euccol(c)) > width && col != 0) {
	    putc('\n', out);
	    row++;
	    col = 0;
	    if (row >= 5)
		return;
	}
	col += n;
	puteucc(c, out);
    }
}

/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	tail - print trailer
 ---------------------------------------------------------------------*/
static void tail(out)
    FILE *out;
{
    for (; row < length; row++)
	putc('\n', out);
    row = -1;
}

/**********************************************************************
 *                                                                    *
 *  Time & Date Routine                                               *
 *                                                                    *
 **********************************************************************/
/*-==(spec:\\:":":\(:\):\[:\]::::prn)==--------------------------------
    TYPE
	local function
    NAME
	strtime - convert date and time to string
    DESCRIPTION
	"strtime()" converts date and time given to
	an 18 character-long string with the form:

	    Sep 16 01:03 1993\0

	Each field has a constant width.  "strtime()" returns a pointer
	to the internally allocated string.
 ---------------------------------------------------------------------*/
static char *strtime(time)
    time_t time;
{
    static struct tm *tm;
    static char buf[20];
    static char *month[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
    };

    tm = localtime(&time);
    sprintf(buf, "%s %2d %2d:%02d %d",
	    month[tm->tm_mon],	/* month:	Jan--Dec*/
	    tm->tm_mday,	/* date:	 0--31 */
	    tm->tm_hour,	/* hour:	 0--23 */
	    tm->tm_min,		/* minute:	00--59 */
	    tm->tm_year + (tm->tm_year < 72 ? 2000 : 1900));
				/* year:	like 1992 */
    return (buf);
}
