/**** << comment & keyword processor >> ****
 *
 *  reg.y
 *                                                  Feb 25 1992
 *                                      mod:        Aug 19 1993
 ************************************************** tonooka ***********/

/*
 * Regular expression matching routines:
 */

/**********************************************************************
 *                                                                    *
 *  Compilation Routines                                              *
 *                                                                    *
 **********************************************************************/
%{
#if !defined lint
static char *sccsid = "@(#)reg.y 2.1 (Y.Tonooka) 11/29/93";
#endif

#include <stdio.h>
#include <ctype.h>
#include "cmt.h"
#include "euc.h"

typedef unsigned char u_char;

static enum {
    NOTLIST,			/* not in list */
    BEGLIST,			/* beginning of list */
    INLIST,			/* in list */
} list;
static union regexp *reg;

static int escape();

static union regexp *makenode();
static union regexp *makechar();
static union regexp *makeclass();
static union regexp *addclass();
static union regexp *negclass();
static union regexp *makeclos();
static union regexp *makealt();
static union regexp *catnode();

%}

%union {
    int chr;
    union regexp *node;
}

%start	expr
%token	<chr>	CHAR CHAR2 CLASS NCLASS DUMMY ERROR
%type	<node>	alt seq closure prim class list

%%

expr	: /* e */		{ reg = makenode(DUMMY); }
	| alt			{ reg = $1; }
	;

alt	: seq
	| alt '|' seq		{ $$ = makealt($1, $3); }
	;

seq	: closure
	| seq closure		{ $$ = catnode($1, $2); }
	;

closure	: prim
	| closure '*'		{ $$ = makeclos('*', $1); }
	| closure '?'		{ $$ = makeclos('?', $1); }
	| closure '+'		{ $$ = makeclos('+', $1); }
	;

prim	: CHAR			{ $$ = makechar($1); }
	| '.'			{ $$ = makenode('.'); }
	| '^'			{ $$ = makenode('^'); }
	| '$'			{ $$ = makenode('$'); }
	| class
	| '(' alt ')'		{ $$ = $2; }
	;

class	: beg		{
			    list = INLIST;
			    $$ = makeclass(CLASS);
			}
	    list ']'	{
			    list = NOTLIST;
			    $$ = $3;
			}
	| beg '^'	{
			    list = INLIST;
			    $$ = makeclass(NCLASS);
			}
	    list ']'	{
			    list = NOTLIST;
			    $$ = $4;
			}
	;

beg	: '['			{ list = BEGLIST; }
	;

list	: '-'			{ $$ = addclass($<node>0, '-', '-'); }
	| ']'			{ $$ = addclass($<node>0, ']', ']'); }
	| elem			{ $$ = $<node>0; }
	| list elem		{ $$ = $<node>0; }
	;

elem	: CHAR			{ (void) addclass($<node>0, $1, $1); }
	| CHAR '-' CHAR		{ (void) addclass($<node>0, $1, $3); }
	;

%%

/*
 * pattern structures
 */
struct node {			/* representitive of all types */
    short id;			/* '^', '$', '.', DUMMY */
    union regexp *next;
};

struct code {
    short id;			/* CHAR, CHAR2 */
    union regexp *next;
    char_t c;			/* char or process code */
};

struct class {
    short id;			/* CLASS, NCLASS */
    union regexp *next;
#if !defined EUC
    unsigned char tab[32];	/* ISO */
#else
    unsigned char tab[16];	/* EUC */
    struct list *list;
#endif
};

#if defined EUC
struct list {
    struct list *next;
    char_t from, to;
};
#endif

struct closure {
    short id;			/* '*', '+', '?' */
    union regexp *next;
    union regexp *expr;
};

struct alter {
    short id;			/* '|' */
    union regexp *next;
    union regexp *expr[2];
};

union regexp {
    struct node node;
    struct code code;
    struct class class;
    struct closure closure;
    struct alter alter;
};

/*
 * static variables
 */
static char_t *sp;
static int onecase;		/* case sensitivity flag */
static char *yyerr;

/*---------------------------------------------------------------------
    NAME
	yylex - lexical analyser
 ---------------------------------------------------------------------*/
int yylex()
{
    register int c;

    if (*sp == '\0')
	return 0;
    c = *sp++;
    if (list != NOTLIST)
	switch (c) {
	case '-':
	case ']':
	    return c;
	case '\\':
	    if ((c = escape()) < 0)
		return ERROR;
	    break;
	case '^':
	    /*
	     * Return '^' only at the beginning of `[...]'.
	     */
	    if (list == BEGLIST)
		return '^';
	    break;
	}
    else
	switch (c) {
	case '\n':
	    return 0;
	case '^': case '$': case '.':
	case '?': case '+': case '*':
	case '[': case '(': case ')':
	case '|':
	    return c;
	case '\\':
	    if ((c = escape()) < 0)
		return ERROR;
	    break;
	}
    yylval.chr = c;
    return CHAR;
}

/*---------------------------------------------------------------------
    NAME
	escape - handle escape sequence
 ---------------------------------------------------------------------*/
static int escape()
{
    register int c;
    register int i;

    if (*sp < 0x80 && isdigit(*sp) && *sp != '8' && *sp != '9')
	for (c = 0, i = 3; i; --i)
	    if (isdigit(*(u_char *) sp) &&
		    *(u_char *) sp != '8' && *(u_char *) sp != '9')
		c *= 8, c += *sp++ - '\0';
	    else
		return -1;
    else {
	switch (*sp) {
	case 't':
	    c = '\t';
	    break;
	case 's':
	    c = ' ';
	    break;
	case 'b':
	    c = '\b';
	    break;
	case 'v':
	    c = '\v';
	    break;
	case 'f':
	    c = '\f';
	    break;
	case 'r':
	    c = '\r';
	    break;
	case '\n':
	    c = '\\';
	    break;
	case '\0':
	    return '\\';
	default:
	    c = *sp;
	    break;
	}
	sp++;
    }
    return c;
}

/*---------------------------------------------------------------------
    NAME
	yyerror
 ---------------------------------------------------------------------*/
yyerror(s)
    char *s;
{
    yyerr = s;
}

/**********************************************************************
 *                                                                    *
 *  Node Creation Functions                                           *
 *                                                                    *
 **********************************************************************/
extern void error();

/*---------------------------------------------------------------------
    NAME
	makenode
 ---------------------------------------------------------------------*/
static union regexp *makenode(id)
    int id;
{
    union regexp *n;

    if ((n = (union regexp *) malloc(sizeof (struct node))) == NULL)
	error("out of memory");
    n->node.next = NULL;
    n->node.id = id;
    return n;
}

/*---------------------------------------------------------------------
    NAME
	makechar
 ---------------------------------------------------------------------*/
static union regexp *makechar(c)
    register int c;
{
    union regexp *n;

    if ((n = (union regexp *) malloc(sizeof (struct code))) == NULL)
	error("out of memory");
    n->code.next = NULL;
    if (!onecase) {
	n->code.id = CHAR;
	n->code.c = c;
    } else {
	n->code.id = CHAR2;
	n->code.c = toupper(c);
    }
    return n;
}

/*---------------------------------------------------------------------
    NAME
	makeclass
 ---------------------------------------------------------------------*/
static union regexp *makeclass(id)
    int id;			/* CLASS, NCLASS */
{
    register unsigned char *p;
    register int i;
    union regexp *n;

    if ((n = (union regexp *) malloc(sizeof (struct class))) == NULL)
	error("out of memory");
    n->class.next = NULL;
    n->class.id = id;
#if !defined EUC
    for (p = n->class.tab, i = 32; i; --i)
	*p++ = 0;
#else
    for (p = n->class.tab, i = 16; i; --i)
	*p++ = 0;
    n->class.list = NULL;
#endif
    return n;
}

/*---------------------------------------------------------------------
    NAME
	addclass
 ---------------------------------------------------------------------*/
static union regexp *addclass(n, from, to)
    union regexp *n;
    int from, to;
{
    register int c;
#if defined EUC
    register struct list *p, *pp;
#endif

#if defined EUC
    if (from > 0x100 || to > 0x100) {
	if (from >= 0x100 && to >= from) {
	    if ((p = (struct list *) malloc(sizeof (struct list)))
		    == NULL)
		error("out of memory");
	    p->next = NULL;
	    p->from = from;
	    p->to = to;
	    if (n->class.list) {
		for (pp = n->class.list; pp->next; pp = pp->next)
		    ;
		pp->next = p;
	    } else
		n->class.list = p;
	}
	return n;
    }
#endif
    for (c = from; c <= to; c++)
	if (!onecase)
	    n->class.tab[c >> 3] |= 1 << (c & 07);
	else {
	    n->class.tab[toupper(c) >> 3] |= 1 << (toupper(c) & 07);
	    n->class.tab[tolower(c) >> 3] |= 1 << (tolower(c) & 07);
	}
    return n;
}

/*---------------------------------------------------------------------
    NAME
	makeclos
 ---------------------------------------------------------------------*/
static union regexp *makeclos(id, e)
    int id;
    union regexp *e;
{
    union regexp *n;

    if ((n = (union regexp *) malloc(sizeof (struct closure))) == NULL)
	error("out of memory");
    n->closure.next = NULL;
    n->closure.id = id;
    n->closure.expr = e;
    return n;
}

/*---------------------------------------------------------------------
    NAME
	makealt
 ---------------------------------------------------------------------*/
static union regexp *makealt(e1, e2)
    union regexp *e1, *e2;
{
    union regexp *n;

    if ((n = (union regexp *) malloc(sizeof (struct alter))) == NULL)
	error("out of memory");
    n->alter.next = NULL;
    n->alter.id = '|';
    n->alter.expr[0] = e1;
    n->alter.expr[1] = e2;
    return n;
}

/*---------------------------------------------------------------------
    NAME
	catnode
 ---------------------------------------------------------------------*/
static union regexp *catnode(parent, child)
    union regexp *parent, *child;
{
    register union regexp *n;
    for (n = parent; n->node.next; n = n->node.next)
	;
    n->node.next = child;
    return parent;
}

/**********************************************************************
 *                                                                    *
 *  Global Routine for Compilation                                    *
 *                                                                    *
 **********************************************************************/
/*---------------------------------------------------------------------
    NAME
	regcomp
    DESCRIPTION
	Compiles regular expression into internal form.
 ---------------------------------------------------------------------*/
union regexp *regcomp(str, oc)
    char_t *str;
    bool oc;			/* non-zero -> case insensitive */
{
    if (str == NULL)
	return NULL;
    sp = str;
    list = NOTLIST;
    yyerr = NULL;
    onecase = oc;
    yyparse();
    return yyerr ? NULL : reg;
}

/**********************************************************************
 *                                                                    *
 *  Matching Routines                                                 *
 *                                                                    *
 **********************************************************************/
static char_t *bol;
static char_t *match();

/*---------------------------------------------------------------------
    NAME
	regmatch
    DESCRIPTION
	If matches, returns pointer to the next character past the end
	of matched string.
 ---------------------------------------------------------------------*/
char_t *regmatch(line, str, exp)
    char_t *line, *str;
    union regexp *exp;
{
    if (str < line)
	return NULL;
    bol = line;
    return match(str, exp);
}

/*---------------------------------------------------------------------
    NAME
	match
 ---------------------------------------------------------------------*/
static char_t *iteration();
static bool isthere();

static char_t *match(s, p)
    register char_t *s;
    register union regexp *p;
{
    register char_t *ss;

    for (; p; p = p->node.next) {
	switch (p->node.id) {
	case CHAR:
	    if (*s++ == p->code.c)
		continue;
	    break;
	case CHAR2:
#if defined EUC
	    if ((*s < 0x80 ? toupper((int) *s++) : *s++) == p->code.c)
		continue;
#else
	    if (toupper(*s++) == p->code.c)
		continue;
#endif
	    break;
	case '^':
	    if (s == bol)
		continue;
	    break;
	case '$':
	    if (*s++ == '\n')
		continue;
	    break;
	case '.':		/* any char but '\n' */
	    if (*s && *s++ != '\n')
		continue;
	    break;
	case CLASS:		/* character class */
	    if (*s && *s != '\n' && isthere((int) *s++, p))
		continue;
	    break;
	case NCLASS:		/* not in character class */
	    if (*s && *s != '\n' && !isthere((int) *s++, p))
		continue;
	    break;
	case '+':		/* at least one match is required */
	    if (ss = match(s, p->closure.expr))
		if (ss = iteration(ss, p))
		    return ss;
	    break;
	case '*':		/* zero or more match */
	    if (ss = iteration(s, p))
		return ss;
	    break;
	case '?':		/* zero or one match */
	    if (ss = match(s, p->closure.expr))
		if (ss = match(ss, p->node.next))
		    return ss;
	    continue;
	case '|':		/* alternation */
	    /*
	     * Match earlier, the winner.
	     */
	    if (ss = match(s, p->alter.expr[0]))
		if (ss = match(ss, p->node.next))
		    return ss;
	    if (ss = match(s, p->alter.expr[1])) {
		s = ss;
		continue;
	    }
	    break;
	case DUMMY:		/* always matches to nothing */
	    continue;
	default:
	    fprintf(stderr, "unknown type %x\n", (int) p->node.id);
	    exit(1);
	    break;
	}
	return NULL;
    }
    return s;
}

/*---------------------------------------------------------------------
    NAME
	iteration
 ---------------------------------------------------------------------*/
static char_t *iteration(str, p)
    char_t *str;
    union regexp *p;
{
    char_t *s;

    if (s = match(str, p->closure.expr))
	if (s = iteration(s, p))
	    return s;
    if (s = match(str, p->closure.next))
	return s;
    return NULL;
}

/*---------------------------------------------------------------------
    NAME
	isthere
 ---------------------------------------------------------------------*/
static bool isthere(c, p)
    register int c;
    union regexp *p;
{
#if defined EUC
    register struct list *k;
#endif

#if defined EUC
    if (c >= 0x100) {
	for (k = p->class.list; k; k = k->next)
	    if (k->from <= c && c <= k->to)
		return 1;
	return 0;
    }
#endif
    return p->class.tab[c >> 3] & 1 << (c & 7);
}

/**********************************************************************
 *                                                                    *
 *  Free Routine                                                      *
 *                                                                    *
 **********************************************************************/
void regfree(p)
    union regexp *p;
{
    union regexp *next;

    for (; p; p = next) {
	next = p->node.next;
	switch (p->node.id) {
#if defined EUC
	case CLASS:
	case NCLASS:
	    if (p->class.list)
		free(p->class.list);
	    break;
#endif
	case '+':
	case '*':
	case '?':
	    regfree(p->closure.expr);
	    break;;
	case '|':
	    regfree(p->alter.expr[0]);
	    regfree(p->alter.expr[1]);
	    break;
	}
	free(p);
    }
}
